from django.apps import apps
from django.conf import settings
import os, inspect
import logging
from ..util import log


TEMPLATE_ATTR_NAME = '_dmp_provider_cache_'

##############################################################
###   Abstract Provider Base

class BaseProvider(object):
    '''
    Abstract base provider class.  An instance is tied to a template at runtime.

    Note that the app can only be inferred for templates in project apps (below settings.BASE_DIR).
    The app has to be inferred because mako creates templates internally during the render runtime,
    and I don't want to hack into Mako internals.

    Always set:
        self.template           Mako template object
        self.options            The combined options from the provider, its supers, and settings.py
        self.template_ext       Template extension.

    If app can be inferred (None if not):
        self.app_config         AppConfig the template resides in, if possible to infer.
        self.template_relpath   Template path, relative to app/templates/, without extension.
                                Usually this is just the template name, but could contain a subdirectory:
                                    homepage/templates/index.html       => "index"
                                    homepage/templates/forms/login.html => "forms/login"
    '''
    DEFAULT_OPTIONS = {
        # the group this provider is part of.  this only matters when
        # the html page limits the providers that will be called with
        # ${ django_mako_plus.links(group="...") }
        'group': 'styles',
        # whether enabled (see "Dev vs. Prod" in the DMP docs)
        'enabled': True,
    }


    @classmethod
    def instance_for_template(cls, template, options):
        '''Returns an instance for the given template'''
        # Mako already caches template objects, so I'm attaching provider instances to templates
        provider_cache = getattr(template, TEMPLATE_ATTR_NAME, None)
        if provider_cache is None:
            provider_cache = {}
            setattr(template, TEMPLATE_ATTR_NAME, provider_cache)
        try:
            return provider_cache[options['_template_cache_key']]
        except KeyError:
            pass
        # not cached yet, so create the object
        instance = cls(template, options)
        if not settings.DEBUG:
            provider_cache[options['_template_cache_key']] = instance
        return instance


    def __init__(self, template, options):
        self.template = template
        self.options = options
        self.app_config = None
        self.template_ext = None
        self.template_relpath = None
        self.template_name = None
        if self.template.filename is not None:
            # try to infer the app
            fn_no_ext, self.template_ext = os.path.splitext(template.filename)
            relpath = os.path.relpath(fn_no_ext, settings.BASE_DIR)
            if not relpath.startswith('..'):  # can't infer reliably outside of project dir
                try:
                    path_parts = os.path.normpath(relpath).split(os.path.sep)
                    self.app_config = apps.get_app_config(path_parts[0])
                    self.template_relpath = '/'.join(path_parts[2:])
                    self.template_name = path_parts[-1]
                except LookupError: # template isn't under an app
                    pass


    def __repr__(self):
        return '<{}{}: {}/{}>'.format(
            self.__class__.__qualname__,
            '' if self.options['enabled'] else ' (disabled)',
            self.app_config.name if self.app_config is not None else 'unknown',
            self.template_relpath if self.template_relpath is not None else self.template,
        )


    @property
    def group(self):
        return self.options['group']


    def start(self, provider_run, data):
        '''
        Called on the *main* template's provider list as the run starts.
        Initialize values in the data dictionary here.
        '''


    def provide(self, provider_run, data):
        '''Called on *each* template's provider list in the chain - use provider_run.write() for content'''


    def finish(self, provider_run, data):
        '''
        Called on the *main* template's provider list as the run finishes
        Finalize values in the data dictionary here.
        '''




#####################################
###  ViewParameter

class ViewParameter(object):
    '''
    A data class that represents a view parameter on a view function.
    An instance of this class is created for each parameter in a view function
    (except the initial request object argument).
    '''
    def __init__(self, name, position, kind, type, default):
        '''
        name:      The name of the parameter.
        position:  The position of this parameter.
        kind:      The kind of argument (positional, keyword, etc.). See inspect module.
        type:      The expected type of this parameter.  Converters use this type to
                   convert urlparam strings to the right type.
        default:   Any default value, specified in function type hints.  If no default is
                   specified in the function, this is `inspect.Parameter.empty`.
        '''
        self.name = name
        self.position = position
        self.kind = kind
        self.type = type
        self.default = default

    def __repr__(self):
        return '<ViewParameter name={}, type={}, default={}>'.format(
            self.name,
            self.type.__qualname__ if self.type is not None else '<not specified>',
            self.default,
        )




import functools

##########################################################
###   An extensible decorator superclass that supports:
###       1. @mydecorator
###       2. @mydecorator()
###       3. @mydecorator(1, 2, 3)
###       4. @mydecorator(a=1, b=2)
###       5. @mydecorator(1, 2, 3, a=1, b=2)
###
###       AND
###
###       A. It works on regular functions.
###       B. It works on unbound class methods.
###       C. It works on bound class methods.
###
###   Author: Conan Albrecht <doconix@gmail.com>
###   Date: 2018-05-05
###
###   See the examples at the end of this file.
###
###   The reason I argue for a metaclass approach:
###
###   When a decorator is hit by python, it either 1) needs to call the decorator function with the
###   other function, or 2) if arguments, run the decorator pre-function, which returns the real decorator,
###   and then #1 with that.
###
###   So in other words, the decorator needs to act like two entirely separate things.  Since a metaclass’
###   job is to create the object, it seems a clean approach to have the metaclass create the right kind
###   of thing needed.
###
###   Since this decision is being made *before* the decorator is created, the decorator class itself
###   can act as just one type of object: a decorator.  The decorator’s `__init__` constructor is super
###   clean because it doesn’t have to check whether we’re in pre-decorator or regular decorator mode.
###   If the decorator constructor is running, it *is decorating* a function.
###

class BaseDecoratorMeta(type):
    '''
    Metaclass that either creates the decorator object or creates a
    factory to create the decorator object.
    '''
    def __call__(self, *args, **kwargs):
        # delegate the creation of instances to a factory
        def factory(func):
            instance = super(BaseDecoratorMeta, self).__call__(func, *args, **kwargs)
            # using custom updated=() because functools updates __dict__.  This replaces
            # instance.decorator_function with the real function--making it point
            # directly at the final function rather than at the next decorator in line
            # (if we have multiple decorators chained together).
            functools.update_wrapper(instance, func, updated=())
            return instance

        # if args has a single callable and kwargs is empty, we'll assume
        # python is calling the decorator: args[0] is the function being decorated.
        # this means the syntax was `@decorator` with no arguments.
        if len(args) == 1 and callable(args[0]) and len(kwargs) == 0:
            func, args = args[0], ()
            return factory(func)

        # if we get here, the syntax was `@decorator(...)` -- with arguments.
        # python hasn't yet called the decorator.  we'll return the factory
        # function so python can call it
        return factory


class BaseDecorator(object, metaclass=BaseDecoratorMeta):
    '''
    A decorator base class that can be called with an arbitrary number of
    arguments and keyword arguments, or with none at all.

        decorator_function: Reference to the next decorator in the chain, or
            to the final function if this is the only/last decorator
        decorator_args: Positional arguments the decorator was called with (if any)
        decorator_kwargs: Named arguments the decorator was called with (if any)

    Get a reference to the real function at the end of the decorator chain with:
        import inspect
        f = inspect.unwrap(self.decorator_function)

        # when there is only one decorator in place, self.decorator_function and
        # inspect.unwrap(self.decorator_function) are the same: the real function.

    If you need to override __init__, follow this pattern:
        class MyDecorator(BaseDecorator):
            def __init__(self, decorator_function, p1, p2=None, *args, **kwargs):
                super().__init__(decorator_function, *args, **kwargs)
                self.p1 = p1
                self.p2 = p2
    '''
    def __init__(self, decorator_function, *decorator_args, **decorator_kwargs):
        self.decorator_function = decorator_function
        self.decorator_args = decorator_args
        self.decorator_kwargs = decorator_kwargs

    def __get__(self, instance, type=None):
        # If we get here, the decorator was placed on a class method.  In this case,
        # decorator_function is actually an unbound function not yet a bound method.
        # Python calls this descriptor when the method is called.
        # When it does so, we need to set the self variable so it gets "bound".
        return functools.partial(self, instance)

    def __call__(self, *args, **kwargs):
        '''Subclasses should override this method'''
        return self.decorator_function(*args, **kwargs)




#######################################################
###   Examples and Simple Testing

if __name__ == '__main__':

    import unittest

    ###  Our Decorator  ###

    class my_decorator(BaseDecorator):
        def __call__(self, *args, **kwargs):
            # do something useful here (the purpose of the decorator)

            # call the decorated function
            result = self.decorator_function(*args, **kwargs)

            # return some things to be tested
            return self, result


    ###  Examples and Tests of Regular Function Decorating  ###

    @my_decorator
    def without_args():
        return 'without_args called'

    @my_decorator()
    def with_empty_args():
        return 'with_empty_args called'

    @my_decorator(1, 2, 3)
    def with_args():
        return 'with_args called'

    @my_decorator(option1='value1', option2=4)
    def with_kwargs():
        return 'with_kwargs called'

    @my_decorator(1, 2, 3, option1='value1', option2=4)
    def with_args_and_kwargs():
        return 'with_args_and_kwargs called'

    class TestFunctionDecorating(unittest.TestCase):
        def test_without_args(self):
            '''Decorating without any arguments or parenthasis'''
            dec, result = without_args()
            self.assertEqual(result, 'without_args called')
            self.assertEqual(dec.decorator_args, ())
            self.assertEqual(dec.decorator_kwargs, {})

        def test_with_empty_args(self):
            '''Decorating with empty arguments'''
            dec, result = with_empty_args()
            self.assertEqual(result, 'with_empty_args called')
            self.assertEqual(dec.decorator_args, ())
            self.assertEqual(dec.decorator_kwargs, {})

        def test_with_args(self):
            '''Decorating with positional arguments'''
            dec, result = with_args()
            self.assertEqual(result, 'with_args called')
            self.assertEqual(dec.decorator_args, (1, 2, 3))
            self.assertEqual(dec.decorator_kwargs, {})

        def test_with_kwargs(self):
            '''Decorating with positional arguments'''
            dec, result = with_kwargs()
            self.assertEqual(result, 'with_kwargs called')
            self.assertEqual(dec.decorator_args, ())
            self.assertEqual(dec.decorator_kwargs, { 'option1':'value1', 'option2':4 })

        def test_with_args_and_kwargs(self):
            '''Decorating with positional arguments'''
            dec, result = with_args_and_kwargs()
            self.assertEqual(result, 'with_args_and_kwargs called')
            self.assertEqual(dec.decorator_args, (1, 2, 3))
            self.assertEqual(dec.decorator_kwargs, { 'option1':'value1', 'option2':4 })


    ###  Examples and Tests of Method Decorating  ###

    class MyTestClass(object):
        @my_decorator
        def without_args(self):
            return 'without_args called'

        @my_decorator()
        def with_empty_args(self):
            return 'with_empty_args called'

        @my_decorator(1, 2, 3)
        def with_args(self):
            return 'with_args called'

        @my_decorator(option1='value1', option2=4)
        def with_kwargs(self):
            return 'with_kwargs called'

        @my_decorator(1, 2, 3, option1='value1', option2=4)
        def with_args_and_kwargs(self):
            return 'with_args_and_kwargs called'

    class TestMethodDecorating(unittest.TestCase):

        def setUp(self):
            self.my_test_class = MyTestClass()

        def test_without_args(self):
            '''Decorating without any arguments or parenthasis'''
            dec, result = self.my_test_class.without_args()
            self.assertEqual(result, 'without_args called')
            self.assertEqual(dec.decorator_args, ())
            self.assertEqual(dec.decorator_kwargs, {})

        def test_with_empty_args(self):
            '''Decorating with empty arguments'''
            dec, result = self.my_test_class.with_empty_args()
            self.assertEqual(result, 'with_empty_args called')
            self.assertEqual(dec.decorator_args, ())
            self.assertEqual(dec.decorator_kwargs, {})

        def test_with_args(self):
            '''Decorating with positional arguments'''
            dec, result = self.my_test_class.with_args()
            self.assertEqual(result, 'with_args called')
            self.assertEqual(dec.decorator_args, (1, 2, 3))
            self.assertEqual(dec.decorator_kwargs, {})

        def test_with_kwargs(self):
            '''Decorating with positional arguments'''
            dec, result = self.my_test_class.with_kwargs()
            self.assertEqual(result, 'with_kwargs called')
            self.assertEqual(dec.decorator_args, ())
            self.assertEqual(dec.decorator_kwargs, { 'option1':'value1', 'option2':4 })

        def test_with_args_and_kwargs(self):
            '''Decorating with positional arguments'''
            dec, result = self.my_test_class.with_args_and_kwargs()
            self.assertEqual(result, 'with_args_and_kwargs called')
            self.assertEqual(dec.decorator_args, (1, 2, 3))
            self.assertEqual(dec.decorator_kwargs, { 'option1':'value1', 'option2':4 })



    class TestInvalidSyntax(unittest.TestCase):

        def test_invalid_syntax(self):
            '''
            Invalid: the only decorator syntax not supported is a single function value
            because python thinks it's the decorated function when it's actually meant
            to be the pre-decorator call.
            '''
            with self.assertRaises(TypeError):
                # we can't tell wether to decorate `sum` or `with_single_callable` here:
                @my_decorator(sum)
                def with_single_callable():
                    print('with_single_callable() called')
                with_single_callable()

    # trigger the testing
    unittest.main()



from django.apps import apps
from django.core.management.commands.startapp import Command as StartAppCommand
from django_mako_plus.management.mixins import DMPCommandMixIn

import os, os.path


NOT_SET = object()


class Command(DMPCommandMixIn, StartAppCommand):
    help = (
        "Creates a DMP app directory structure for the given app name in "
        "the current directory or optionally in the given directory."
    )
    requires_system_checks = False

    def add_arguments(self, parser):
        super().add_arguments(parser)
        self.get_action_by_dest(parser, 'template').default = NOT_SET


    def handle(self, *args, **options):
        dmp = apps.get_app_config('django_mako_plus')
        if options.get('template') is NOT_SET:
            # set the template to a DMP app
            options['template'] = 'http://cdn.rawgit.com/doconix/django-mako-plus/master/app_template.zip'
            # attempt to use a local DMP install instead of the online repo as specified above
            dmp_dir = dmp.path
            if dmp_dir:
                template_dir = os.path.join(dmp_dir, 'app_template')
                if os.path.exists(template_dir):
                    options['template'] = template_dir

        # ensure we have the extensions we need
        options['extensions'] = list(set(options.get('extensions') + [ 'py', 'htm', 'html' ]))

        # call the super
        StartAppCommand.handle(self, *args, **options)

        self.message("""App {name} created successfully!

What's next?
    1. Add your new app to the list in settings.py:
        INSTALLED_APPS = [
            ...
            '{name}',
        ]
    2. python manage.py runserver
    3. Take a browser to http://localhost:8000/

""".format(name=options.get('name')))



from django.core.exceptions import ImproperlyConfigured
from django.conf import settings
from django.forms.utils import flatatt
import logging
import os
import os.path
import posixpath
from ..util import crc32, getdefaultattr
from ..util import log
from .base import BaseProvider


#####################################################
###   LinkProvider abstract base class

class LinkProvider(BaseProvider):
    '''
    Renders links like <link> and <script> based on the name of the template
    and supertemplates.
    '''
    DEFAULT_OPTIONS = {
        # explicitly sets the path to search for - if this filepath exists, DMP
        # includes a link to it in the template. globs are not supported because this
        # should resolve to one exact file. possible values:
        #   1. None: a default path is used, such as "{app}/{subdir}/{filename.ext}", prefixed
        #      with the static root at production; see subclasses for their default filenames.
        #   2. function, lambda, or other callable: called as func(provider) and
        #      should return a string
        #   3. str: used directly
        'filepath': None,

        # explicitly sets the link to be inserted into the template (when filepath exists).
        # possible values:
        #   1. None: a default path is used, such as "<script src="..."></script>, prefixed
        #      with the static url at production; see subclasses for the default link format.
        #   2. function, lambda, or other callable: called as func(provider) and
        #      should return a string
        #   3. str: inserted directly into the template
        'link': None,

        # if a template is rendered more than once in a request, should we link more than once?
        # defaults are: css=False, js=True, bundled_js=False
        'skip_duplicates': False,
    }

    def __init__(self, template, options):
        super().__init__(template, options)
        self.filepath = os.path.join(
            settings.BASE_DIR if settings.DEBUG else settings.STATIC_ROOT,
            self.build_source_filepath()
        )
        # file time and version hash
        try:
            self.mtime = int(os.stat(self.filepath).st_mtime)
            # version_id combines current time and the CRC32 checksum of file bytes
            self.version_id = (self.mtime << 32) | crc32(self.filepath)
        except FileNotFoundError:
            self.mtime = 0
            self.version_id = 0
        if log.isEnabledFor(logging.DEBUG):
            log.debug('%s created for %s: [%s]', repr(self), self.filepath, 'will link' if self.mtime > 0 else 'will skip nonexistent file')


    ### Source Filepath Building Methods ###

    def build_source_filepath(self):
        # if defined in settings, run the function or return the string
        if self.options['filepath'] is not None:
            return self.options['filepath'](self) if callable(self.options['filepath']) else self.options['filepath']
        # build the default
        if self.app_config is None:
            log.warn('{} skipped: template %s not in project subdir and `targetpath` not in settings', (self.__class__.__qualname__, self.template_relpath))
        return self.build_default_filepath()

    def build_default_filepath(self):
        raise ImproperlyConfigured('{} must set `filepath` in options (or a subclass can override build_default_filepath).'.format(self.__class__.__qualname__))


    ### Target Link Building Methods ###

    def build_target_link(self, provider_run, data):
        # if defined in settings, run the function or return the string
        if self.options['link'] is not None:
            return self.options['link'](self) if callable(self.options['link']) else self.options['link']
        # build the default
        if self.app_config is None:
            log.warn('{} skipped: template %s not in project subdir and `targetpath` not in settings', (self.__class__.__qualname__, self.template_relpath))
        return self.build_default_link(provider_run, data)

    def build_default_link(self, provider_run, data):
        raise ImproperlyConfigured('{} must set `link` in options (or a subclass can override build_default_link).'.format(self.__class__.__qualname__))


    ### Provider Run Methods ###

    def start(self, provider_run, data):
        # add a set to the request (fallback to provider_run if request is None) for skipping duplicates
        if self.options['skip_duplicates']:
            data['seen'] = getdefaultattr(
                provider_run.request.dmp if provider_run.request is not None else provider_run,
                '_LinkProvider_Filename_Cache_',
                factory=set,
            )
        # enabled providers in the chain go here
        data['enabled'] = []

    def provide(self, provider_run, data):
        filepath = self.filepath
        # delaying printing of tag to finish() because the JsContextProvider delays and this must go after it
        # short circuit if the file for this provider doesn't exist
        if self.mtime == 0:
            return
        # short circut if we're skipping duplicates and we've already seen this one
        if self.options['skip_duplicates']:
            if filepath in data['seen']:
                if log.isEnabledFor(logging.DEBUG):
                    log.debug('%s skipped duplicate %s', repr(self), self.filepath)
                return
            data['seen'].add(filepath)
        # if we get here, this provider is enabled, so add it to the list
        if log.isEnabledFor(logging.DEBUG):
            log.debug('%s linking %s', repr(self), self.filepath)
        data['enabled'].append(self)

    def finish(self, provider_run, data):
        for provider in data['enabled']:
            provider_run.write(provider.build_target_link(provider_run, data))


#####################################
###  CssLinkProvider

class CssLinkProvider(LinkProvider):
    '''Generates a CSS <link>'''
    DEFAULT_OPTIONS = {
        'group': 'styles',
        # if a template is rendered more than once in a request, we usually don't
        # need to include the css again.
        'skip_duplicates': True,
    }

    def build_default_filepath(self):
        return os.path.join(
            self.app_config.name,
            'styles',
            self.template_relpath + '.css',
        )

    def build_default_link(self, provider_run, data):
        attrs = {}
        attrs['rel'] = 'stylesheet'
        attrs["data-context"] = provider_run.uid
        attrs["href"] ="{}?{:x}".format(
            # posixpath because URLs use forward slash
            posixpath.join(
                settings.STATIC_URL,
                self.app_config.name,
                'styles',
                self.template_relpath.replace(os.path.sep, '/') + '.css',
            ),
            self.version_id,
        )
        return '<link{} />'.format(flatatt(attrs))


##############################################
###   JsLinkProvider

class JsLinkProvider(LinkProvider):
    '''Generates a JS <script>.'''
    DEFAULT_OPTIONS = {
        'group': 'scripts',
        # if a template is rendered more than once in a request, we should link each one
        # so the script runs again each time the template runs
        'skip_duplicates': False,
        # whether to create an async script tag
        'async': False,
    }

    def build_default_filepath(self):
        return os.path.join(
            self.app_config.name,
            'scripts',
            self.template_relpath + '.js',
        )

    def build_default_link(self, provider_run, data):
        attrs = {}
        attrs["data-context"] = provider_run.uid
        attrs["src"] ="{}?{:x}".format(
            # posixpath because URLs use forward slash
            posixpath.join(
                settings.STATIC_URL,
                self.app_config.name,
                'scripts',
                self.template_relpath.replace(os.path.sep, '/') + '.js',
            ),
            self.version_id,
        )
        if self.options['async']:
            attrs['async'] = "async"
        return '<script{}></script>'.format(flatatt(attrs))



import os, os.path



#########################################
###  Mixin for all DMP commands

class DMPCommandMixIn(object):
    '''Some extra SWAG that all DMP commands get'''
    # needs to be true so Django initializes urls.py (which registers the dmp apps)
    requires_system_checks = True

    def add_arguments(self, parser):
        super().add_arguments(parser)

        # django also provides a verbosity parameter
        # these two are just convenience params to it
        parser.add_argument(
            '--verbose',
            action='store_true',
            dest='verbose',
            default=False,
            help='Set verbosity to level 3 (see --verbosity).',
        )
        parser.add_argument(
            '--quiet',
            action='store_true',
            dest='quiet',
            default=False,
            help='Set verbosity to level 0, which silences all messages (see --verbosity).',
        )


    def get_action_by_dest(self, parser, dest):
        '''Retrieves the given parser action object by its dest= attribute'''
        for action in parser._actions:
            if action.dest == dest:
                return action
        return None


    def execute(self, *args, **options):
        '''Placing this in execute because then subclass handle() don't have to call super'''
        if options['verbose']:
            options['verbosity'] = 3
        if options['quiet']:
            options['verbosity'] = 0
        self.verbosity = options.get('verbosity', 1)
        super().execute(*args, **options)


    def get_dmp_path(self):
        '''Returns the absolute path to DMP.  Apps do not have to be loaded yet'''
        return os.path.dirname(os.path.dirname(__file__))


    def message(self, msg='', level=1, tab=0):
        '''Print a message to the console'''
        if self.verbosity >= level:
            self.stdout.write('{}{}'.format('    ' * tab, msg))



from django.apps import apps
from django.core.management.commands.startproject import Command as StartProjectCommand
from django_mako_plus.management.mixins import DMPCommandMixIn

import os, os.path


NOT_SET = object()


class Command(DMPCommandMixIn, StartProjectCommand):
    help = (
        "Creates a DMP project directory structure for the given project "
        "name in the current directory or optionally in the given directory."
    )
    requires_system_checks = False

    def add_arguments(self, parser):
        super().add_arguments(parser)
        self.get_action_by_dest(parser, 'template').default = NOT_SET

    def handle(self, *args, **options):
        if options.get('template') is NOT_SET:
            # set the template to a DMP app
            options['template'] = 'http://cdn.rawgit.com/doconix/django-mako-plus/master/project_template.zip'
            # attempt to use a local DMP install instead of the online repo as specified above
            # this should work unless the installation type is not normal
            template_dir = os.path.join(self.get_dmp_path(), 'project_template')
            if os.path.exists(template_dir):
                options['template'] = template_dir

        # call the super
        StartProjectCommand.handle(self, *args, **options)

        self.message("""Project {name} created successfully!

What's next?
    1. cd {name}
    2. python3 manage.py dmp_startapp homepage

""".format(name=options['name']))



from django.apps import apps
from django.utils.module_loading import import_string
from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from django_mako_plus.template import create_mako_context
from django_mako_plus.provider.runner import ProviderRun
from django_mako_plus.management.mixins import DMPCommandMixIn
from mako.template import Template as MakoTemplate

import os
import os.path
from collections import OrderedDict



class Command(DMPCommandMixIn, BaseCommand):
    help = 'Removes compiled template cache folders in your DMP-enabled app directories.'
    can_import_settings = True


    def add_arguments(self, parser):
        super().add_arguments(parser)

        parser.add_argument(
            '--overwrite',
            action='store_true',
            dest='overwrite',
            default=False,
            help='Overwrite existing __entry__.js if necessary'
        )
        parser.add_argument(
            '--single',
            type=str,
            metavar='FILENAME',
            help='Instead of per-app entry files, create a single file that includes the JS for all listed apps'
        )
        parser.add_argument(
            'appname',
            type=str,
            nargs='*',
            help='The name of one or more DMP apps. If omitted, all DMP apps are processed'
        )


    def handle(self, *args, **options):
        dmp = apps.get_app_config('django_mako_plus')
        self.options = options
        WebpackProviderRun.initialize_providers()

        # ensure we have a base directory
        try:
            if not os.path.isdir(os.path.abspath(settings.BASE_DIR)):
                raise CommandError('Your settings.py BASE_DIR setting is not a valid directory.  Please check your settings.py file for the BASE_DIR variable.')
        except AttributeError:
            raise CommandError('Your settings.py file is missing the BASE_DIR setting.')

        # the apps to process
        enapps = []
        for appname in options.get('appname'):
            enapps.append(apps.get_app_config(appname))
        if len(enapps) == 0:
            enapps = dmp.get_registered_apps()

        # main runner for per-app files
        if options.get('single') is None:
            for app in enapps:
                self.message('Searching `{}` app...'.format(app.name))
                filename = os.path.join(app.path, 'scripts', '__entry__.js')
                self.create_entry_file(filename, self.generate_script_map(app), [ app ])

        # main runner for one sitewide file
        else:
            script_map = {}
            for app in enapps:
                self.message('Searching `{}` app...'.format(app.name))
                script_map.update(self.generate_script_map(app))
            self.create_entry_file(options.get('single'), script_map, enapps)


    def create_entry_file(self, filename, script_map, enapps):
        '''Creates an entry file for the given script map'''
        if len(script_map) == 0:
            return

        # delete previous file if it exists, and ensure the target directory is there
        if os.path.exists(filename):
            if self.options.get('overwrite'):
                os.remove(filename)
            else:
                raise CommandError('Refusing to destroy existing file: {} (use --overwrite option or remove the file)'.format(filename))
        if not os.path.exists(os.path.dirname(filename)):
            os.makedirs(os.path.dirname(filename))

        # write the entry file
        self.message('Creating {}'.format(os.path.relpath(filename, settings.BASE_DIR)))
        template = MakoTemplate('''
<%! from datetime import datetime %>
<%! import sys %>
<%! import os %>

// Generated on ${ datetime.now().strftime('%Y-%m-%d %H:%M') } by `${ ' '.join(sys.argv) }`
// Contains links for ${ 'app' if len(enapps) == 1 else 'apps' }: ${ ', '.join(sorted([ a.name for a in enapps ])) }

(context => {
    DMP_CONTEXT.loadBundle({
      %for (app, template), script_paths in script_map.items():
        "${ app }/${ template }": function() {
          %for path in script_paths:
            require("./${ os.path.relpath(path, os.path.dirname(filename)) }");
          %endfor
        },
      %endfor
    });
})(DMP_CONTEXT.get());
''')
        with open(filename, 'w') as fout:
            fout.write(template.render(
                enapps=enapps,
                script_map=script_map,
                filename=filename,
            ).strip())


    def generate_script_map(self, config):
        '''
        Maps templates in this app to their scripts.  This function deep searches
        app/templates/* for the templates of this app.  Returns the following
        dictionary with absolute paths:

        {
            ( 'appname', 'template1' ): [ '/abs/path/to/scripts/template1.js', '/abs/path/to/scripts/supertemplate1.js' ],
            ( 'appname', 'template2' ): [ '/abs/path/to/scripts/template2.js', '/abs/path/to/scripts/supertemplate2.js', '/abs/path/to/scripts/supersuper2.js' ],
            ...
        }

        Any files or subdirectories starting with double-underscores (e.g. __dmpcache__) are skipped.
        '''
        script_map = OrderedDict()
        template_root = os.path.join(os.path.relpath(config.path, settings.BASE_DIR), 'templates')
        def recurse(folder):
            subdirs = []
            if os.path.exists(folder):
                for filename in os.listdir(folder):
                    if filename.startswith('__'):
                        continue
                    filerel = os.path.join(folder, filename)
                    if os.path.isdir(filerel):
                        subdirs.append(filerel)

                    elif os.path.isfile(filerel):
                        template_name = os.path.relpath(filerel, template_root)
                        scripts = self.template_scripts(config, template_name)
                        key = ( config.name, os.path.splitext(template_name)[0] )
                        self.message('Found template: {}; static files: {}'.format(key, scripts), 3)
                        script_map[key] = scripts

            for subdir in subdirs:
                recurse(subdir)

        recurse(template_root)
        return script_map


    def template_scripts(self, config, template_name):
        '''
        Returns a list of scripts used by the given template object AND its ancestors.

        This runs a ProviderRun on the given template (as if it were being displayed).
        This allows the WEBPACK_PROVIDERS to provide the JS files to us.

        For this algorithm to work, the providers must extend static_links.LinkProvider
        because that class creates the 'enabled' list of provider instances, and each
        provider instance has a url.
        '''
        dmp = apps.get_app_config('django_mako_plus')
        template_obj = dmp.engine.get_template_loader(config, create=True).get_mako_template(template_name, force=True)
        mako_context = create_mako_context(template_obj)
        inner_run = WebpackProviderRun(mako_context['self'])
        inner_run.run()
        scripts = []
        for data in inner_run.column_data:
            for provider in data.get('enabled', []):    # if file doesn't exist, the provider won't be enabled
                if hasattr(provider, 'filepath'):       # providers used to collect for webpack must have a .filepath property
                    scripts.append(provider.filepath)   # the absolute path to the file (see providers/static_links.py)
        return scripts



###############################################################################
###   Specialized provider run for the above management command

class WebpackProviderRun(ProviderRun):
    SETTINGS_KEY = 'WEBPACK_PROVIDERS'

    def get_template_inheritance(self, tself):
        '''
        Normally, this returns a list of the template inheritance of tself, starting with the oldest ancestor.
        But for the webpack one, we just want the template itself, without any ancestors.
        '''
        return [ tself.template ]



from django.http import HttpResponseRedirect, HttpResponsePermanentRedirect
from django.http import Http404

from .http import HttpResponseJavascriptRedirect, REDIRECT_HEADER_KEY


###############################################################
###   Exceptions used to direct the controller



class BaseRedirectException(Exception):
    '''
    Superclass of DMP redirect exceptions
    '''


class InternalRedirectException(BaseRedirectException):
    '''
    View functions can throw this exception to indicate that a new view
    should be called by the HtmlPageServer.  The current view function
    will end immediately, and processing will be passed to the new view function.
    '''
    def __init__(self, redirect_module, redirect_function):
        '''
        Indicates the new view to be called.  The view should be given relative to the project root.
        The parameters should be strings, not the actual module or function reference.

        Both parameters should be strings, example:

            raise InternalRedirectException('homepage.views.someview', 'process_request')
        '''
        super().__init__()
        self.redirect_module = redirect_module
        self.redirect_function = redirect_function


class RedirectException(BaseRedirectException):
    '''
    Immediately stops processing of a view function or template and redirects to the given page
    using the standard 302 response status header.

    After the redirect_to parameter, you can use any of the normal HttpResponse constructor arguments.

    A custom header is set in the response.  This allows middleware, your web server, or
    calling JS code to adjust the redirect if needed.
    '''
    def __init__(self, redirect_to, *args, **kwargs):
        super().__init__()
        self.redirect_to = redirect_to
        self.args = args
        self.kwargs = kwargs

    def get_response(self, request, *args, **kwargs):
        '''Returns the redirect response for this exception.'''
        # normal process
        response = HttpResponseRedirect(self.redirect_to)
        response[REDIRECT_HEADER_KEY] = self.redirect_to
        return response


class PermanentRedirectException(RedirectException):
    '''
    Immediately stops processing of a view function or template and redirects to the given page
    using the standard 301 response status header.

    After the redirect_to parameter, you can use any of the normal HttpResponse constructor arguments.

    A custom header is set in the response.  This allows middleware, your web server, or
    calling JS code to adjust the redirect if needed.

    '''
    def get_response(self, request):
        '''Returns the redirect response for this exception.'''
        response = HttpResponsePermanentRedirect(self.redirect_to, *self.args, **self.kwargs)
        response[REDIRECT_HEADER_KEY] = self.redirect_to
        return response


class JavascriptRedirectException(RedirectException):
    '''
    Immediately stops processing of a view function or template and redirects to the given page.

    Sends a regular HTTP 200 OK response that contains Javascript to
    redirect the browser:

        <script>window.location.href="...";</script>.

    If redirect_to is empty, it redirects to the current location (essentially refreshing
    the current page):

        <script>window.location.href=window.location.href;</script>.

    Normally, redirecting should be done via HTTP 302 rather than Javascript.
    Use this class when your only choice is through Javascript.

    For example, suppose you need to redirect the top-level page from an Ajax response.
    Ajax redirects normally only redirects the Ajax itself (not the page that initiated the call),
    and this default behavior is usually what is needed.  However, there are instances when the
    entire page must be redirected, even if the call is Ajax-based.

    After the redirect_to parameter, you can use any of the normal HttpResponse constructor arguments.

    If you need to omit the surrounding <script> tags, send "include_script_tag=False" to
    the constructor. One use case for omitting the tags is when the caller is a
    JQuery $.script() ajax call.

    A custom header is set in the response.  This allows middleware, your web server, or
    calling JS code to adjust the redirect if needed.

    Note that this method doesn't use the <meta> tag or Refresh header method because
    they aren't predictable within Ajax (for example, JQuery seems to ignore them).
    '''
    def get_response(self, request):
        '''Returns the redirect response for this exception.'''
        # the redirect key is already placed in the response by HttpResponseJavascriptRedirect
        return HttpResponseJavascriptRedirect(self.redirect_to, *self.args, **self.kwargs)





###############################################
###  Exceptions for the conversion process


class ConverterHttp404(Http404):
    '''
    Customization of Django's Http404 for parameter conversion errors.
    This is raised when a converter function for a type raises a ValueError.
    '''
    def __init__(self, value, parameter, *args, **kwargs):
        self.value = value
        self.parameter = parameter
        super().__init__(*args, **kwargs)


class ConverterException(Exception):
    '''
    Raised when an unknown converter exception occurs.
    '''
    def __init__(self, value, parameter, *args, **kwargs):
        self.value = value
        self.parameter = parameter
        super().__init__(*args, **kwargs)



from django.apps import apps
from django.core.exceptions import ImproperlyConfigured


import inspect
import sys



class ConverterFunctionInfo(object):
    '''Holds information about a converter function'''
    def __init__(self, convert_func, convert_type, source_order):
        self.convert_func = convert_func
        self.convert_type = convert_type
        self.source_order = source_order
        self.sort_key = 0


    def prepare_sort_key(self):
        '''
        Triggered by view_function._sort_converters when our sort key should be created.
        This can't be called in the constructor because Django models might not be ready yet.
        '''
        if isinstance(self.convert_type, str):
            try:
                app_name, model_name = self.convert_type.split('.')
            except ValueError:
                raise ImproperlyConfigured('"{}" is not a valid converter type. String-based converter types must be specified in "app.Model" format.'.format(self.convert_type))
            try:
                self.convert_type = apps.get_model(app_name, model_name)
            except LookupError as e:
                raise ImproperlyConfigured('"{}" is not a valid model name. {}'.format(self.convert_type, e))

        # we reverse sort by ( len(mro), source code order ) so subclasses match first
        # on same types, last declared method sorts first
        self.sort_key = ( -1 * len(inspect.getmro(self.convert_type)), -1 * self.source_order )



from django.apps import apps
from django.conf import settings
from django.utils.module_loading import import_string
from collections import namedtuple
import io
import inspect
from uuid import uuid1
from ..template import template_inheritance
from ..util import qualified_name, b58enc


# I can't keep the options inside the provider class itself because a given class
# can be listed more than once in settings.py (with different options).
# So instead I keep a list of the classes and their options here.
ProviderClassInfo = namedtuple("ProviderClassInfo", [ 'cls', 'options' ])

####################################################
###   Main runner for providers

class ProviderRun(object):
    '''A run through the providers for tself and its ancestors'''
    SETTINGS_KEY = 'CONTENT_PROVIDERS'
    CONTENT_PROVIDERS = []

    @classmethod
    def initialize_providers(cls):
        '''Initializes the providers (called from dmp app ready())'''
        dmp = apps.get_app_config('django_mako_plus')
        # regular content providers
        cls.CONTENT_PROVIDERS = []
        for provider_settings in dmp.options[cls.SETTINGS_KEY]:
            # import the class for this provider
            assert 'provider' in provider_settings, "Invalid entry in settings.py: CONTENT_PROVIDERS item must have 'provider' key"
            provider_cls = import_string(provider_settings['provider'])
            # combine options from all of its bases, then from settings.py
            options = {}
            for base in reversed(inspect.getmro(provider_cls)):
                options.update(getattr(base, 'DEFAULT_OPTIONS', {}))
            options.update(provider_settings)
            # add to the list
            if options['enabled']:
                # the index in the provider list is needed because a given class
                # can be listed more than once in settings.py (with different options)
                options['_template_cache_key'] = '_{}_{}_'.format(qualified_name(provider_cls), len(cls.CONTENT_PROVIDERS))
                cls.CONTENT_PROVIDERS.append(ProviderClassInfo(provider_cls, options))


    def __init__(self, tself, group=None):
        '''
        tself:              `self` object from a Mako template (available during rendering).
        group:              provider group to include (defaults to all groups if None)
        '''
        # a unique context id for this run
        self.uid = b58enc(uuid1().int)
        self.request = tself.context.get('request')
        self.context = tself.context
        self.buffer = io.StringIO()

        # Create a table of providers for each template in the ancestry:
        #
        #     base.htm,      [ JsLinkProvider1, CssLinkProvider1, ... ]
        #        |
        #     app_base.htm,  [ JsLinkProvider2, CssLinkProvider2, ... ]
        #        |
        #     index.html,    [ JsLinkProvider3, CssLinkProvider3, ... ]
        self.template_providers = []
        for template in self.get_template_inheritance(tself):
            providers = []
            for pci in self.CONTENT_PROVIDERS:
                provider = pci.cls.instance_for_template(template, pci.options)
                if group is None or provider.group == group:
                    providers.append(provider)
            self.template_providers.append(providers)

        # Column-specific data dictionaries are maintained as the template providers run
        # (one per provider column).  These allow the provider instances of a given
        # column to share data if needed.
        #
        #      column_data = [ { col 1 }      , { col 2 }      , ... ]
        self.column_data = [ {} for pf in self.CONTENT_PROVIDERS ]


    def get_template_inheritance(self, tself):
        '''Returns a list of the template inheritance of tself, starting with the oldest ancestor'''
        return reversed(list(template_inheritance(tself)))


    def run(self):
        '''Performs the run through the templates and their providers'''
        # start() on tself (the last template in the list)
        for providers in self.template_providers[-1:]:
            for provider, data in zip(providers, self.column_data):
                provider.start(self, data)
        # provide() on the all provider lists in the chain
        for providers in self.template_providers:
            for provider, data in zip(providers, self.column_data):
                provider.provide(self, data)
        # finish() on tself (the last template in the list)
        for providers in self.template_providers[-1:]:
            for provider, data in zip(providers, self.column_data):
                provider.finish(self, data)


    def write(self, content):
        '''Provider instances use this to write to the buffer'''
        self.buffer.write(content)
        if settings.DEBUG:
            self.buffer.write('\n')


    def getvalue(self):
        '''Returns the buffer string'''
        return self.buffer.getvalue()



from django import template
from django.utils.safestring import mark_safe
from ..convenience import render_template

register = template.Library()

###########################################
###  The tags in this file are DJANGO
###  tags, not Mako ones.

@register.simple_tag(takes_context=True)
def dmp_include(context, app, template_name, def_name=None):
    '''
    Includes a DMP (Mako) template into a normal django template.

    In other words, this is used whenyou have a Django-style template and
    need to include another template within it -- but that template is written
    in DMP (Mako) syntax rather than Django syntax.
    '''
    return mark_safe(render_template(
        request=context.get('request'),
        app=app,
        template_name=template_name,
        context=context.flatten(),
        def_name=def_name
    ))



from django.apps import apps, AppConfig
from django.core.exceptions import ImproperlyConfigured
from django.template import TemplateDoesNotExist
from django.template.backends.base import BaseEngine
from django.utils.module_loading import import_string
from mako.template import Template

from .template import MakoTemplateLoader, MakoTemplateAdapter

import itertools
import os
import os.path
import re


# Following Django's lead, hard coding the CSRF processor
BUILTIN_CONTEXT_PROCESSORS = (
    'django_mako_plus.context_processors.csrf',
)

# regex to split the template name in get_template()
# "myapp/mytemplate.html#myblock" gives groups:
# ('myapp', '/mytemplate.html', 'mytemplate.html', '#myblock', 'myblock')
RE_TEMPLATE_NAME = re.compile('([^/?#]*)?(/([^#]*))?(#(.*))?')


#########################################################
###   The main engine


class MakoTemplates(BaseEngine):
    '''
    The primary Mako interface that plugs into the Django templating system.
    This is referenced in settings.py -> TEMPLATES.
    '''
    def __init__(self, params):
        '''Constructor'''
        # ensure DMP is listed as an app (common install error)
        try:
            self.dmp = apps.get_app_config('django_mako_plus')
        except LookupError:
            raise ImproperlyConfigured("`django_mako_plus` must be listed in INSTALLED_APPS before it can be used")

        # cache for our template loaders
        self.template_loaders = {}

        # set up the context processors
        context_processors = []
        for processor in itertools.chain(BUILTIN_CONTEXT_PROCESSORS, self.dmp.options['CONTEXT_PROCESSORS']):
            context_processors.append(import_string(processor))
        self.template_context_processors = tuple(context_processors)

        # super constructor
        params.pop('OPTIONS', None)   # the super doesn't like OPTIONS in there
        super(MakoTemplates, self).__init__(params)


    def from_string(self, template_code):
        '''
        Compiles a template from the given string.
        This is one of the required methods of Django template engines.
        '''
        dmp = apps.get_app_config('django_mako_plus')
        mako_template = Template(template_code, imports=dmp.template_imports, input_encoding=dmp.options['DEFAULT_TEMPLATE_ENCODING'])
        return MakoTemplateAdapter(mako_template)


    def get_template(self, template_name):
        '''
        Retrieves a template object from the pattern "app_name/template.html".
        This is one of the required methods of Django template engines.

        Because DMP templates are always app-specific (Django only searches
        a global set of directories), the template_name MUST be in the format:
        "app_name/template.html" (even on Windows).  DMP splits the template_name
        string on the slash to get the app name and template name.

        Template rendering can be limited to a specific def/block within the template
        by specifying `#def_name`, e.g. `myapp/mytemplate.html#myblockname`.
        '''
        dmp = apps.get_app_config('django_mako_plus')
        match = RE_TEMPLATE_NAME.match(template_name)
        if match is None or match.group(1) is None or match.group(3) is None:
            raise TemplateDoesNotExist('Invalid template_name format for a DMP template.  This method requires that the template name be in app_name/template.html format (separated by slash).')
        if not dmp.is_registered_app(match.group(1)):
            raise TemplateDoesNotExist('Not a DMP app, so deferring to other template engines for this template')
        return self.get_template_loader(match.group(1)).get_template(match.group(3), def_name=match.group(5))


    def get_template_loader(self, app, subdir='templates', create=False):
        '''
        Returns a template loader object for the given app name in the given subdir.
        For example, get_template_loader('homepage', 'styles') will return
        a loader for the styles/ directory in the homepage app.

        The app parameter can be either an app name or an AppConfig instance.
        The subdir parameter is normally 'templates', 'scripts', or 'styles',
        but it can be any subdirectory name of the given app.

        Normally, you should not have to call this method.  Django automatically
        generates two shortcut functions for every DMP-registered apps,
        and these shortcut functions are the preferred way to render templates.

        This method is useful when you want a custom template loader to a directory
        that does not conform to the app_dir/templates/* pattern.

        If the loader is not found in the DMP cache, one of two things occur:
          1. If create=True, it is created automatically and returned.  This overrides
             the need to register the app as a DMP app.
          2. If create=False, a TemplateDoesNotExist is raised.  This is the default
             behavior.
        '''
        # ensure we have an AppConfig
        if app is None:
            raise TemplateDoesNotExist("Cannot locate loader when app is None")
        if not isinstance(app, AppConfig):
            app = apps.get_app_config(app)

        # get the loader with the path of this app+subdir
        path = os.path.join(app.path, subdir)

        # if create=False, the loader must already exist in the cache
        if not create:
            dmp = apps.get_app_config('django_mako_plus')
            if not dmp.is_registered_app(app):
                raise ValueError("{} is not registered with DMP [hint: check urls.py for include('django_mako_plus.urls')].".format(app))

        # return the template by path
        return self.get_template_loader_for_path(path, use_cache=True)


    def get_template_loader_for_path(self, path, use_cache=True):
        '''
        Returns a template loader object for the given directory path.
        For example, get_template_loader('/var/mytemplates/') will return
        a loader for that specific directory.

        Normally, you should not have to call this method.  Django automatically
        adds request.dmp.render() and request.dmp.render_to_string() on each
        request.

        This method is useful when you want a custom template loader for a specific
        directory that may be outside your project directory or that is otherwise
        not contained in a normal Django app.  If the directory is inside an app,
        call get_template_loader() instead.

        Unless use_cache=False, this method caches template loaders in the DMP
        cache for later use.
        '''
        # get from the cache if we are able
        if use_cache:
            try:
                return self.template_loaders[path]
            except KeyError:
                pass  # not there, so we'll create

        # create the loader
        loader = MakoTemplateLoader(path, None)

        # cache if we are allowed
        if use_cache:
            self.template_loaders[path] = loader

        # return
        return loader



from django.apps import apps
from django.core.exceptions import ViewDoesNotExist
from django.http import Http404
from django.urls import ResolverMatch
from django.urls.exceptions import Resolver404
try:
    from django.urls import re_path              # Django 2.x
    from django.urls import URLPattern
    from django.urls import include
    from django.urls.resolvers import RegexPattern
except ImportError:
    from django.conf.urls import url as re_path  # Django 1.x
    from django.urls import RegexURLPattern as URLPattern
    from django.conf.urls import include
    RegexPattern = None

from ..util import merge_dicts, log
from .data import RoutingData
from .decorators import RequestViewWrapper

from collections import namedtuple



##################################################
###  DMP-style resolver for an app


def app_resolver(app_name=None, pattern_kwargs=None, name=None):
    '''
    Registers the given app_name with DMP and adds convention-based
    url patterns for it.

    This function is meant to be called in a project's urls.py.
    '''
    urlconf = URLConf(app_name, pattern_kwargs)
    resolver = re_path(
        '^{}/?'.format(app_name) if app_name is not None else '',
        include(urlconf),
        name=urlconf.app_name,
    )
    # this next line is a workaround for Django's URLResolver class not having
    # a `name` attribute, which is expected in Django's technical_404.html.
    resolver.name = getattr(resolver, 'name', name or app_name)
    return resolver


class URLConf(object):
    '''
    Mimics a urls.py module for an app by holding a list of url patterns.
    This is run through Django's include() function to create a resolver.

    Call `app_resolver()` to instantiate this class.  Do not create it directly.
    '''
    def __init__(self, app_name=None, pattern_kwargs=None):
        self._app_name = app_name
        self.app_name = self._app_name or '<default app>'
        self.name = self.app_name
        dmp = apps.get_app_config('django_mako_plus')
        dmp.register_app(self._app_name)
        self.urlpatterns = dmp_paths_for_app(self._app_name, pattern_kwargs, self.app_name)


def dmp_paths_for_app(app_name, pattern_kwargs=None, pretty_app_name=None):
    '''Utility function that creates the default patterns for an app'''
    dmp = apps.get_app_config('django_mako_plus')
    # Because these patterns are subpatterns within the app's resolver,
    # we don't include the /app/ in the pattern -- it's already been
    # handled by the app's resolver.
    #
    # Also note how the each pattern below defines the four kwargs--
    # either as 1) a regex named group or 2) in kwargs.
    return [
        # page.function/urlparams
        dmp_path(
            r'^(?P<dmp_page>[_a-zA-Z0-9\-]+)\.(?P<dmp_function>[_a-zA-Z0-9\.\-]+)/(?P<dmp_urlparams>.+?)/?$',
            merge_dicts({
                'dmp_app': app_name or dmp.options['DEFAULT_APP'],
            }, pattern_kwargs),
            'DMP /{}/page.function/urlparams'.format(pretty_app_name),
            app_name,
        ),

        # page.function
        dmp_path(
            r'^(?P<dmp_page>[_a-zA-Z0-9\-]+)\.(?P<dmp_function>[_a-zA-Z0-9\.\-]+)/?$',
            merge_dicts({
                'dmp_app': app_name or dmp.options['DEFAULT_APP'],
                'dmp_urlparams': '',
            }, pattern_kwargs),
            'DMP /{}/page.function'.format(pretty_app_name),
            app_name,
        ),

        # page/urlparams
        dmp_path(
            r'^(?P<dmp_page>[_a-zA-Z0-9\-]+)/(?P<dmp_urlparams>.+?)/?$',
            merge_dicts({
                'dmp_app': app_name or dmp.options['DEFAULT_APP'],
                'dmp_function': 'process_request',
            }, pattern_kwargs),
            'DMP /{}/page/urlparams'.format(pretty_app_name),
            app_name,
        ),

        # page
        dmp_path(
            r'^(?P<dmp_page>[_a-zA-Z0-9\-]+)/?$',
            merge_dicts({
                'dmp_app': app_name or dmp.options['DEFAULT_APP'],
                'dmp_function': 'process_request',
                'dmp_urlparams': '',
            }, pattern_kwargs),
            'DMP /{}/page'.format(pretty_app_name),
            app_name,
        ),

        # empty
        dmp_path(
            r'^$',
            merge_dicts({
                'dmp_app': app_name or dmp.options['DEFAULT_APP'],
                'dmp_function': 'process_request',
                'dmp_urlparams': '',
                'dmp_page': dmp.options['DEFAULT_PAGE'],
            }, pattern_kwargs),
            'DMP /{}'.format(pretty_app_name),
            app_name,
        ),
    ]




#############################################
###  DMP-style pattern


def dmp_path(regex, kwargs=None, name=None, app_name=None):
    '''
    Creates a DMP-style, convention-based pattern that resolves
    to various view functions based on the 'dmp_page' value.

    The following should exist as 1) regex named groups or
    2) items in the kwargs dict:
        dmp_app         Should resolve to a name in INSTALLED_APPS.
                        If missing, defaults to DEFAULT_APP.
        dmp_page        The page name, which should resolve to a module:
                        project_dir/{dmp_app}/views/{dmp_page}.py
                        If missing, defaults to DEFAULT_PAGE.
        dmp_function    The function name (or View class name) within the module.
                        If missing, defaults to 'process_request'
        dmp_urlparams   The urlparams string to parse.
                        If missing, defaults to ''.

    The reason for this convenience function is to be similar to
    Django functions like url(), re_path(), and path().
    '''
    return PagePattern(regex, kwargs, name, app_name)


class PagePattern(URLPattern):
    '''
    Creates a DMP-style, convention-based pattern that resolves
    to various view functions based on the 'dmp_page' value.
    '''
    def __init__(self, regex, default_args=None, name=None, app_name=None):
        self.dmp = apps.get_app_config('django_mako_plus')
        if app_name or self.dmp.options['DEFAULT_APP']:
            self.dmp.register_app(app_name)
        default_args = merge_dicts({ 'dmp_app': app_name or self.dmp.options['DEFAULT_APP'] }, default_args)
        if isinstance(regex, str) and RegexPattern is not None:
            regex = RegexPattern(regex, name=name, is_endpoint=True)
        # this is a bit of a hack, but the super constructor needs
        # a view function. Our resolve() function ignores this view function
        # so we'll just creeate a no-op placeholder so super is happy.
        def no_op_view(request):
            raise Http404()
        super().__init__(regex, no_op_view, default_args, name=name)


    def resolve(self, path):
        '''
        Different from Django, this method matches by /app/page/ convention
        using its pattern.  The pattern should create keyword arguments for
        dmp_app, dmp_page.
        '''
        match = super().resolve(path)
        if match:
            try:
                routing_data = RoutingData(
                    match.kwargs.pop('dmp_app', None) or self.dmp.options['DEFAULT_APP'],
                    match.kwargs.pop('dmp_page', None) or self.dmp.options['DEFAULT_PAGE'],
                    match.kwargs.pop('dmp_function', None) or 'process_request',
                    match.kwargs.pop('dmp_urlparams', '').strip(),
                )
                return ResolverMatch(
                    RequestViewWrapper(routing_data),
                    match.args,
                    match.kwargs,
                    match.url_name,
                )
            except ViewDoesNotExist as vdne:
                # we had a pattern match, but we couldn't get a callable using kwargs from the pattern
                # create a "pattern" so the programmer can see what happened
                # this is a hack, but the resolver error page doesn't give other options.
                # the sad face is to catch the dev's attention in Django's printout
                msg = "◉︵◉ Pattern matched, but discovery failed: {}".format(vdne)
                log.debug("%s %s", match.url_name, msg)
                raise Resolver404({
                    # this is a bit convoluted, but it makes the PatternStub work with Django 1.x and 2.x
                    'tried': [[ PatternStub(match.url_name, msg, PatternStub(match.url_name, msg, None)) ]],
                    'path': path,
                })



from collections import namedtuple
PatternStub = namedtuple('PatternStub', [ 'name', 'pattern', 'regex' ])



###########################################################################################
###   Converter of Base10 (decimal) to Base58
###   Ambiguous chars not used: 0, O, I, and l
###   This uses the same alphabet as bitcoin.

BASE58CHARS = "123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz"
BASE58INDEX = { ch: i for i, ch in enumerate(BASE58CHARS) }

def b58enc(uid):
    '''Encodes a UID to an 11-length string, encoded using base58 url-safe alphabet'''
    # note: i tested a buffer array too, but string concat was 2x faster
    if not isinstance(uid, int):
        raise Base58Error('Invalid integer: {}'.format(uid))
    if uid == 0:
        return BASE58CHARS[0]
    enc_uid = ""
    while uid:
        uid, r = divmod(uid, 58)
        enc_uid = BASE58CHARS[r] + enc_uid
    return enc_uid

def b58dec(enc_uid):
    '''Decodes a UID from base58, url-safe alphabet back to int.'''
    if isinstance(enc_uid, str):
        pass
    elif isinstance(enc_uid, bytes):
        enc_uid = enc_uid.decode('utf8')
    else:
        raise Base58Error('Cannot decode this type: {}'.format(enc_uid))
    uid = 0
    try:
        for i, ch in enumerate(enc_uid):
            uid = (uid * 58) + BASE58INDEX[ch]
    except KeyError:
        raise Base58Error('Invalid character: "{}" ("{}", index 5)'.format(ch, enc_uid, i))
    return uid

class Base58Error(Exception):
    pass



from django.apps import apps
from django.core.management.base import BaseCommand, CommandError
from django.conf import settings
from django_mako_plus.management.mixins import DMPCommandMixIn

import os, os.path, shutil
import sys
import argparse

# this command was placed here in Sept 2018 to help users know about the change.
# it can probably be removed sometime in Summer, 2019.


class Command(DMPCommandMixIn, BaseCommand):
    help = 'Message to inform users of the change back to dmp_* commands.'

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(dest='all', nargs=argparse.REMAINDER, help='Wildcard to catch all remaining arguments')

    def handle(self, *args, **options):
        try:
            pos = sys.argv.index('dmp')  # should be 1
            guess = 'Our guess at the right command is:\n\n    {}'.format(' '.join(
                sys.argv[:pos] + \
                [ sys.argv[pos] + '_' + sys.argv[pos+1] ] + \
                sys.argv[pos+2:]
            ))
        except:  # `dmp` not there, or no subcommand
            guess = ''
        raise CommandError('''

DMP command usage changed in v5.6: `manage.py dmp *` commands are now `manage.py dmp_*`.

As much as we liked the former syntax, it had to be tied to Django internals. It broke
whenever Django changed its internal command structure. Apologies for the change.

{}
        '''.format(guess))




# public items in this package
from .parameter import ViewParameter
from .decorators import parameter_converter
from .base import ParameterConverter


# import the default converters
# this must come at the end of the file so view_function above is loaded
# it doesn't matter what's imported -- the file just needs to load
from .converters import __name__ as _



from django.core.exceptions import ImproperlyConfigured

from ..util import log
from ..exceptions import ConverterHttp404, ConverterException
from .parameter import ViewParameter
from .info import ConverterFunctionInfo

import logging
import inspect
from operator import attrgetter
import functools


class ParameterConverter(object):
    '''
    Converts parameters using functions registered to types they convert.

    To create a converter for a new type, simply create a function
    that is decorated with @parameter_converter

    Customizing Parameter Conversion:

    The primary way to customize conversion is to create new @parameter_converter
    functions in your codebase.  This allows you to add new type converters.
    You can even create functions for types already handled by the built-in
    converters.  Your functions will override the built-in ones.

    If you need to customize more than just a few types, override:

        convert_value(): where individual parameters are converted.

        __call__(): the controller that iterates the parameter values
        and calls convert_value().
    '''

    # the registry of converters (populated by the @converter_function decorator)
    converters = []
    # this variable prevents sorting until Django is ready (because during sorting
    # we switch "myapp.MyModel" to the actual model instance)
    _sorting_enabled = False

    def __init__(self, view_function):
        self.view_function = view_function

        # inspect the parameters on the function (or functions if a class-based view)
        self.view_parameters = {}
        view_class = getattr(self.view_function, 'view_class', None)
        if view_class is None:  # regular view function
            self.view_parameters[None] = self._collect_parameters(self.view_function)

        else:  # class-based view
            for http_mthd in view_class.http_method_names:
                func = getattr(view_class, http_mthd, None)
                if func is not None:
                    self.view_parameters[http_mthd] = self._collect_parameters(func, True)
            # Django's View class aliases head to get using this logic
            if 'get' in self.view_parameters and 'head' not in self.view_parameters:
                self.view_parameters['head'] = self.view_parameters['get']


    def _collect_parameters(self, func, class_based=False):
        func_parameters = list(inspect.signature(func).parameters.values())
        # when using class-based views, methods that have decorators might be partials,
        # which makes it difficult to know whether the `self` parameter is present.
        # this heuristic is the best way I can figure out to skip the self parameter if there.
        if class_based and len(func_parameters) > 0 and func_parameters[0].name == 'self':
            func_parameters = func_parameters[1:]

        params = []
        for i, p in enumerate(func_parameters):
            params.append(ViewParameter(
                name=p.name,
                position=i,
                kind=p.kind,
                type=p.annotation,
                default=p.default,
            ))
        return tuple(params)


    @classmethod
    def _register_converter(cls, conv_func, conv_type):
        '''Triggered by the @converter_function decorator'''
        cls.converters.append(ConverterFunctionInfo(conv_func, conv_type, len(cls.converters)))
        cls._sort_converters()


    @classmethod
    def _sort_converters(cls, app_ready=False):
        '''Sorts the converter functions'''
        # app_ready is True when called from DMP's AppConfig.ready()
        # we can't sort before then because models aren't ready
        cls._sorting_enabled = cls._sorting_enabled or app_ready
        if cls._sorting_enabled:
            for converter in cls.converters:
                converter.prepare_sort_key()
            cls.converters.sort(key=attrgetter('sort_key'))


    def convert_parameters(self, request, *args, **kwargs):
        '''
        Iterates the urlparams and converts them according to the
        type hints in the current view function.  This is the primary
        function of the class.
        '''
        args = list(args)
        urlparam_i = 0

        parameters = self.view_parameters.get(request.method.lower()) or self.view_parameters.get(None)
        if parameters is not None:
            # add urlparams into the arguments and convert the values
            for parameter_i, parameter in enumerate(parameters):
                # skip request object, *args, **kwargs
                if parameter_i == 0 or parameter.kind is inspect.Parameter.VAR_POSITIONAL or parameter.kind is inspect.Parameter.VAR_KEYWORD:
                    pass
                # value in kwargs?
                elif parameter.name in kwargs:
                    kwargs[parameter.name] = self.convert_value(kwargs[parameter.name], parameter, request)
                # value in args?
                elif parameter_i - 1 < len(args):
                    args[parameter_i - 1] = self.convert_value(args[parameter_i - 1], parameter, request)
                # urlparam value?
                elif urlparam_i < len(request.dmp.urlparams):
                    kwargs[parameter.name] = self.convert_value(request.dmp.urlparams[urlparam_i], parameter, request)
                    urlparam_i += 1
                # can we assign a default value?
                elif parameter.default is not inspect.Parameter.empty:
                    kwargs[parameter.name] = self.convert_value(parameter.default, parameter, request)
                # fallback is None
                else:
                    kwargs[parameter.name] = self.convert_value(None, parameter, request)

        return args, kwargs


    def convert_value(self, value, parameter, request):
        '''
        Converts a parameter value in the view function call.

            value:      value from request.dmp.urlparams to convert
                        The value will always be a string, even if empty '' (never None).

            parameter:  an instance of django_mako_plus.ViewParameter that holds this parameter's
                        name, type, position, etc.

            request:    the current request object.

        "converter functions" register with this class using the @parameter_converter
        decorator.  See converters.py for the built-in converters.

        This function goes through the list of registered converter functions,
        selects the most-specific one that matches the parameter.type, and
        calls it to convert the value.

        If the converter function raises a ValueError, it is caught and
        switched to an Http404 to tell the browser that the requested URL
        doesn't resolve to a page.

        Other useful exceptions that converter functions can raise are:

            RedirectException: redirects the browser (see DMP docs)
            InternalRedirectException: redirects processing internally (see DMP docs)
            Http404: returns a Django Http404 response
        '''
        try:
            # we don't convert anything without type hints
            if parameter.type is inspect.Parameter.empty:
                if log.isEnabledFor(logging.DEBUG):
                    log.debug('skipping conversion of parameter `%s` because it has no type hint', parameter.name)
                return value

            # find the converter method for this type
            # I'm iterating through the list to find the most specific match first
            # The list is sorted by specificity so subclasses come before their superclasses
            for ci in self.converters:
                if issubclass(parameter.type, ci.convert_type):
                    if log.isEnabledFor(logging.DEBUG):
                        log.debug('converting parameter `%s` using %s', parameter.name, ci.convert_func)
                    return ci.convert_func(value, parameter)

            # if we get here, there wasn't a converter or this type
            raise ImproperlyConfigured(message='No parameter converter exists for type: {}. Do you need to add an @parameter_converter function for the type?'.format(parameter.type))

        except ValueError as e:
            log.info('ValueError raised during conversion of parameter %s (%s): %s', parameter.position, parameter.name, e)
            raise ConverterHttp404(value, parameter, 'A parameter could not be converted - see the logs for more detail') from e

        except Exception as e:
            log.info('Exception raised during conversion of parameter %s (%s): %s', parameter.position, parameter.name, e)
            raise ConverterException(value, parameter, 'A parameter could not be converted - see the logs for more detail') from e



#!/usr/bin/env python3
from django.core import management
import os.path
import sys
import functools

__doc__ = '''
Starts a new DMP-style project.  This is the DMP equivalent of "django-admin.py startproject".

Example:

    django_mako_plus dmp_startproject [project name]

if the above doesn't work, try:

    python -m django_mako_plus dmp_startproject [project name]

Background: The dmp_startproject command has a bit of a chicken-and-egg problem.
It creates the project, but the command isn't available until the project is created
and DMP is in INSTALLED_APPS.  Can't create until it is created...
Django solves this with django-admin.py, a global Python script that can be
executed directly from the command line.
'''

DMP_MANAGEMENT_PATH = os.path.join(os.path.dirname(__file__), 'management')



def main():
    # Django is hard coded to return only its own commands when a project doesn't
    # exist yet.  Since I want DMP to be able to create projects, I'm monkey-patching
    # Django's get_commands() function so DMP gets added.  This is the least-offensive
    # way I could see to do this since Django really isn't built to allow other commands
    # pre-project.  This only happens when `django_mako_plus` is run directly and not
    # when `manage.py` or `django-admin.py` are run.
    orig_get_commands = management.get_commands
    @functools.lru_cache(maxsize=None)
    def new_get_commands():
        commands = {}
        commands.update(orig_get_commands())
        commands.update({ name: 'django_mako_plus' for name in management.find_commands(DMP_MANAGEMENT_PATH) })
        return commands
    management.get_commands = new_get_commands

    # mimic the code in django-admin.py
    management.execute_from_command_line()


## runner!
if __name__ == '__main__':
    main()



from django.apps import apps
from django.utils.html import conditional_escape
from django.utils.encoding import force_text

from mako.lexer import Lexer
from mako import parsetree, ast

from ..util import log
from ..tags import is_autoescape


###########################################################
###  DMP hook for expression filters. This allows DMP
###  to filter expressions ${...} after all other filters
###  have been run.
###
###  Currently, the use of this hook is HTML autoescaping.
###  Django autoescapes by default, while Mako does not.
###  DMP injects autoescaping to be consistent with Django.
###

MAKO_ESCAPE_REPLACEMENTS = {
    'h': 'django.utils.html.escape',  # uses Django's escape rather than Mako's, which works better with marks
}

class DMPLexer(Lexer):
    '''
    Subclass of Mako's Lexer, which is used during compilation of
    templates.  This subclass injects ExpressionPostProcessor
    as the final filter on every expression.  Overriding append_node()
    is a hack, but it's the only way I can find to hook into Mako's
    compile process without modifying Mako directly.
    '''
    def append_node(self, nodecls, *args, **kwargs):
        # fyi, this method runs on template compilation (not on template render)
        if nodecls == parsetree.Expression:
            # when an Expression, args[1] is a comma-separated string of filters
            # parse the filters and make any DMP replacements for them
            try:
                # this is Mako's ast, not the python one
                filters = [ MAKO_ESCAPE_REPLACEMENTS.get(f, f) for f in ast.ArgumentList(args[1]).args ]
            except Exception as e:
                log.warning('An error occurred when compiling the filters on an expression; allowing through so Mako can handle it (%s)', e)
                filters = []
            extra = {}  # extra info sent to the expression processor

            # if we have the 'n' filter, send that to the expression processor
            if 'n' in filters:
                extra['n_filter_on'] = True

            # add the expression processor as the last filter to be run
            # then recreate the args tuple
            filters.append('django_mako_plus.ExpressionPostProcessor(self' + \
                           (", extra={}".format(extra) if len(extra) > 0 else '') + \
                           ')')
            args = args[:1] + (','.join(filters),) + args[2:]
        return super().append_node(nodecls, *args, **kwargs)


# this is used read-only, so it can be in __init__ signature
EMPTY_DICT = {}

class ExpressionPostProcessor(object):
    '''
    Object that is called as the final filter on every template
    expression ${...}.

    Right now this object does autoescaping. However, it is placed
    on *every* expression so we have a hook for future post-processing
    of expressions.

    See the creation of this object in DMPLexer above for more info.
    '''
    def __init__(self, tself, extra=EMPTY_DICT):
        # check whether it's on for this block
        self.html_escape = is_autoescape(tself.context)
        # the 'n' filter turns off our normal html escaping
        if extra.get('n_filter_on', False):
            self.html_escape = False
        # mark_safe() is handled in Django's conditional_escape(), so no need to deal with it
        # check the global setting
        dmp = apps.get_app_config('django_mako_plus')
        if not dmp.options['AUTOESCAPE']:
            self.html_escape = False

    def __call__(self, text):
        '''
        Mako calls this after evaluating the expression and applying
        all other filters.

        Right now this html escapes the expression, unless autoescape
        is toggled off.
        '''
        # we apply this across the board, even if the `n` filter is present because
        # DMP always creates unicode (see adapter.py where render_unicode() is used)
        text = force_text(text)

        # html encoding
        if self.html_escape:
            text = conditional_escape(text)  # internally, this honors mark_safe()

        return text



from django.apps import apps
from django.core.exceptions import ImproperlyConfigured, ViewDoesNotExist
from django.http import HttpResponse, StreamingHttpResponse, Http404, HttpResponseServerError

from ..decorators import BaseDecorator
from ..signals import dmp_signal_post_process_request, dmp_signal_pre_process_request, dmp_signal_internal_redirect_exception, dmp_signal_redirect_exception
from ..util import import_qualified, log
from ..exceptions import InternalRedirectException, RedirectException

import inspect
import sys
import functools

# key we use to attach the converter to the view function (also see discover.py)
CONVERTER_ATTRIBUTE_NAME = 'parameter_converter'


##########################################
###   View-function decorator

class view_function(BaseDecorator):
    '''
    A decorator to signify which view functions are "callable" by web browsers.
    and to convert parameters using type hints, if provided.

    All endpoint functions, such as process_request, must be decorated as:

        @view_function
        function process_request(request):
            ...

    Or:

        @view_function(...)
        function process_request(request):
            ...
    '''
    # singleton set of decorated functions
    DECORATED_FUNCTIONS = set()


    def __init__(self, decorator_function, *args, **kwargs):
        '''Create a new wrapper around the decorated function'''
        super().__init__(decorator_function, *args, **kwargs)
        real_func = inspect.unwrap(decorator_function)

        # flag the function as an endpoint. doing it on the actual function because
        # we don't know the order of decorators on the function. order only matters if
        # the other decorators don't use @wraps correctly .in that case, @view_function
        # will put DECORATED_KEY on the decorator function rather than the real function.
        # but even that is fine *as long as @view_function is listed first*.
        self.DECORATED_FUNCTIONS.add(real_func)


    @classmethod
    def is_decorated(cls, f):
        '''Returns True if the given function is decorated with @view_function'''
        real_func = inspect.unwrap(f)
        return real_func in cls.DECORATED_FUNCTIONS



#############################################
###  Per-request wrapper for view functions

class RequestViewWrapper(object):
    '''
    A wrapper for the view function, created for each request. This is different than
    @view_function above because @view_function wraps the function one time -- like
    a normal decorator.  This decorator is used like a normal function (no @ syntax)
    and is placed *each* time a request comes through.  It must be created per request
    because we need to store the RoutingInfo object in it (which is different each request).

    Back story: Django creates its own ResolverMatch object during url resolution -- replacing
    the one we create in resolver.py. This makes it difficult to send the RoutingData object,
    through to the rest of the request.  Since the only thing Django keeps from our ResolverMatch
    is the function + args + kwargs, we need to stash the object in one of those.

    This per-request decorator also does parameter conversion, triggers signals,
    and loop RedirectExceptions.
    '''
    # FYI, not using BaseDecorator super because its metaclass
    # expects the function in the constructor arguments.

    def __init__(self, routing_data):
        self.routing_data = routing_data
        # take name and attributes of the view function
        functools.update_wrapper(self, self.routing_data.callable)


    def __call__(self, request, *args, **kwargs):
        log.info('%s', self.routing_data)
        dmp = apps.get_app_config('django_mako_plus')

        # the middleware attaches the routing data to request.dmp, but the
        # middleware is optional. let's attach it here again for those not
        # using the middleware
        request.dmp = self.routing_data

        # an outer try that catches the redirect exceptions
        try:

            # convert the parameters (the converter is placed on the func in discover.py)
            converter = getattr(self.routing_data.callable, CONVERTER_ATTRIBUTE_NAME, None)
            if converter is not None:
                args, kwargs = converter.convert_parameters(request, *args, **kwargs)

            # send the pre-signal
            if dmp.options['SIGNALS']:
                for receiver, ret_response in dmp_signal_pre_process_request.send(sender=sys.modules[__name__], request=request, view_args=args, view_kwargs=kwargs):
                    if isinstance(ret_response, (HttpResponse, StreamingHttpResponse)):
                        return ret_response

            # call the view function
            response = self.routing_data.callable(request, *args, **kwargs)
            if not isinstance(response, (HttpResponse, StreamingHttpResponse)):
                log.info('%s failed to return an HttpResponse (or the post-signal overwrote it).  Returning 500 error.', self.routing_data.callable)
                return HttpResponseServerError('Invalid response received from server.')

            # send the post-signal
            if dmp.options['SIGNALS']:
                for receiver, ret_response in dmp_signal_post_process_request.send(sender=sys.modules[__name__], request=request, response=response, view_args=args, view_kwargs=kwargs):
                    if ret_response is not None:
                        response = ret_response # sets it to the last non-None in the signal receiver chain

            return response

        except InternalRedirectException as ivr:
            # send the signal
            if dmp.options['SIGNALS']:
                dmp_signal_internal_redirect_exception.send(sender=sys.modules[__name__], request=request, exc=ivr)
            # update the RoutingData object
            request.dmp.module = ivr.redirect_module
            request.dmp.function = ivr.redirect_function
            try:
                request.dmp.callable = getattr(import_qualified(request.dmp.module), request.dmp.function)
            except (ImportError, AttributeError):
                log.info('could not fulfill InternalViewRedirect because %s.%s does not exist.', request.dmp.module, request.dmp.function)
                raise Http404()
            # recurse with this routing data
            log.info('received an InternalViewRedirect to %s.%s', request.dmp.module, request.dmp.function)
            return RequestViewWrapper(self.routing_data)(request, *args, **kwargs)

        except RedirectException as e: # redirect to another page
            log.info('view %s.%s redirected processing to %s', request.dmp.module, request.dmp.function, e.redirect_to)
            # send the signal
            if dmp.options['SIGNALS']:
                dmp_signal_redirect_exception.send(sender=sys.modules[__name__], request=request, exc=e)
            # send the browser the redirect command
            return e.get_response(request)

        # the code should never get here



from mako.runtime import supports_caller

###
###  Mako-style tags that DMP provides
###


#########################################################
###  Autoescaping toggling

# attaching to `caller_stack` because it's the same object
# throughout rendering of a template inheritance
AUTOESCAPE_KEY = '__dmp_autoescape'

def is_autoescape(context):
    return bool(getattr(context.caller_stack, AUTOESCAPE_KEY, True))

def toggle_autoescape(context, escape_on=True):
    previous = is_autoescape(context)
    setattr(context.caller_stack, AUTOESCAPE_KEY, escape_on)
    try:
        context['caller'].body()
    finally:
        setattr(context.caller_stack, AUTOESCAPE_KEY, previous)


@supports_caller
def autoescape_on(context):
    '''
    Explicitly turns autoescaping on for a given block within a template,
    (individual filters can still override with ${ somevar | n }).

    Example use in template:
        <%namespace name="dmp" module="django_mako_plus.tags"/>

        <%dmp:autoescape_on>
            ${ somevar } will be autoescaped.
        </%dmp:autoescape_on>
    '''
    toggle_autoescape(context, True)
    return ''


@supports_caller
def autoescape_off(context):
    '''
    Explicitly turns autoescaping off for a given block within a template,
    (individual filters can still override with ${ somevar | h }).

    Example use in template:
        <%namespace name="dmp" module="django_mako_plus.tags"/>

        <%dmp:autoescape>
            ${ somevar } will not be autoescaped.
        </%dmp:autoescape>
    '''
    toggle_autoescape(context, False)
    return ''



from django.apps import apps
from django.core.management.base import CommandError
from django.contrib.staticfiles.management.commands.collectstatic import Command as CollectStaticCommand
from django.conf import settings
from django_mako_plus.management.mixins import DMPCommandMixIn

import os, os.path, shutil, fnmatch

try:
    from rjsmin import jsmin
except ImportError:
    jsmin = None
try:
    from rcssmin import cssmin
except ImportError:
    cssmin = None



TYPE_DIRECTORY = 0       # file is a directory
TYPE_FILE = 1            # file is a regular file


class Rule(object):
    def __init__(self, pattern, level, filetype, score):
        self.pattern = pattern
        self.level = level
        self.filetype = filetype
        self.score = score

    def match(self, fname, flevel, ftype):
        '''Returns the result score if the file matches this rule'''
        # if filetype is the same
        # and level isn't set or level is the same
        # and pattern matche the filename
        if self.filetype == ftype and (self.level is None or self.level == flevel) and fnmatch.fnmatch(fname, self.pattern):
            return self.score
        return 0





class Command(DMPCommandMixIn, CollectStaticCommand):
    help = 'Collects static files, such as media, scripts, and styles, to a common directory root. This is done to prepare for deployment.'
    can_import_settings = True

    def add_arguments(self, parser):
        super().add_arguments(parser)

        parser.add_argument(
            '--overwrite',
            action='store_true',
            dest='overwrite',
            default=False,
            help='Overwrite existing files in the directory when necessary.'
        )
        parser.add_argument(
            '--no-minify',
            action='store_true',
            dest='no_minify',
            help='Do not minify *.css with rcssmin and *.js with rjsmin.',
        )
        parser.add_argument(
            '--include-dir',
            action='append',
            dest='include_dir',
            help='Include directories matching this pattern.  Unix-style wildcards are acceptable, such as "*partial*".  This option can be specified more than once.'
        )
        parser.add_argument(
            '--include-file',
            action='append',
            dest='include_file',
            help='Include files matching this pattern.  Unix-style wildcards are acceptable, such as "*.txt".  This option can be specified more than once.'
        )
        parser.add_argument(
            '--skip-dir',
            action='append',
            dest='skip_dir',
            help='Skip directories matching this pattern.  Unix-style wildcards are acceptable, such as "*partial*".  This option can be specified more than once.'
        )
        parser.add_argument(
            '--skip-file',
            action='append',
            dest='skip_file',
            help='Skip files matching this pattern.  Unix-style wildcards are acceptable, such as "*.txt".  This option can be specified more than once.'
        )


    def handle(self, *args, **options):
        dmp = apps.get_app_config('django_mako_plus')
        self.options = options

        # ensure we have a base directory
        try:
            if not os.path.isdir(os.path.abspath(settings.BASE_DIR)):
                raise CommandError('Your settings.py BASE_DIR setting is not a valid directory.  Please check your settings.py file for the BASE_DIR variable.')
        except AttributeError as e:
            print(e)
            raise CommandError('Your settings.py file is missing the BASE_DIR setting. Aborting app creation.')

        # get the destination directory, and ensure it doesn't already exist
        # if dest_root starts with a /, it is an absolute directory
        # if dest_root doesn't start with a /, it goes relative to the BASE_DIR
        try:
            dest_root = os.path.join(os.path.abspath(settings.BASE_DIR), settings.STATIC_ROOT)
            if os.path.exists(dest_root) and not self.options.get('overwrite'):
                raise CommandError('The destination directory for static files (%s) already exists. Please delete it or run this command with the --overwrite option.' % dest_root)
        except AttributeError:
            raise CommandError('Your settings.py file is missing the STATIC_ROOT setting. Exiting without collecting the static files.')

        # create the directory - we assume it either doesn't exist, or the --overwrite is specified
        if not os.path.isdir(dest_root):
            os.makedirs(dest_root)

        # set up the rules
        self.rules = self.create_rules()

        # go through the DMP apps and collect the static files
        for config in dmp.get_registered_apps():
            self.message('Processing app {}'.format(config.name), 1)
            self.copy_dir(config.path, os.path.abspath(os.path.join(dest_root, config.name)))


    def create_rules(self):
        '''Adds rules for the command line options'''
        dmp = apps.get_app_config('django_mako_plus')
        # the default
        rules = [
            # files are included by default
            Rule('*',                                    level=None, filetype=TYPE_FILE,      score=1),
            # files at the app level are skipped
            Rule('*',                                    level=0,    filetype=TYPE_FILE,      score=-2),
            # directories are recursed by default
            Rule('*',                                    level=None, filetype=TYPE_DIRECTORY, score=1),
            # directories at the app level are skipped
            Rule('*',                                    level=0,    filetype=TYPE_DIRECTORY, score=-2),

            # media, scripts, styles directories are what we want to copy
            Rule('media',                                level=0,    filetype=TYPE_DIRECTORY, score=6),
            Rule('scripts',                              level=0,    filetype=TYPE_DIRECTORY, score=6),
            Rule('styles',                               level=0,    filetype=TYPE_DIRECTORY, score=6),

            # ignore the template cache directories
            Rule(dmp.options['TEMPLATES_CACHE_DIR'],     level=None, filetype=TYPE_DIRECTORY, score=-3),
            # ignore python cache directories
            Rule('__pycache__',                          level=None, filetype=TYPE_DIRECTORY, score=-3),
            # ignore compiled python files
            Rule('*.pyc',                                level=None, filetype=TYPE_FILE,      score=-3),
        ]
        # include rules have score of 50 because they trump all initial rules
        for pattern in (self.options.get('include_dir') or []):
            self.message('Setting rule - recurse directories: {}'.format(pattern), 1)
            rules.append(Rule(pattern, level=None, filetype=TYPE_DIRECTORY, score=50))
        for pattern in (self.options.get('include_file') or []):
            self.message('Setting rule - include files: {}'.format(pattern), 1)
            rules.append(Rule(pattern, level=None, filetype=TYPE_FILE, score=50))
        # skip rules have score of 100 because they trump everything, including the includes from the command line
        for pattern in (self.options.get('skip_dir') or []):
            self.message('Setting rule - skip directories: {}'.format(pattern), 1)
            rules.append(Rule(pattern, level=None, filetype=TYPE_DIRECTORY, score=-100))
        for pattern in (self.options.get('skip_file') or []):
            self.message('Setting rule - skip files: {}'.format(pattern), 1)
            rules.append(Rule(pattern, level=None, filetype=TYPE_FILE, score=-100))
        return rules


    def copy_dir(self, source, dest, level=0):
        '''Copies the static files from one directory to another.  If this command is run, we assume the user wants to overwrite any existing files.'''
        encoding = settings.DEFAULT_CHARSET or 'utf8'
        msglevel = 2 if level == 0 else 3
        self.message('Directory: {}'.format(source), msglevel, level)

        # create a directory for this app
        if not os.path.exists(dest):
            self.message('Creating directory: {}'.format(dest), msglevel, level+1)
            os.mkdir(dest)

        # go through the files in this app
        for fname in os.listdir(source):
            source_path = os.path.join(source, fname)
            dest_path = os.path.join(dest, fname)
            ext = os.path.splitext(fname)[1].lower()

            # get the score for this file
            score = 0
            for rule in self.rules:
                score += rule.match(fname, level, TYPE_DIRECTORY if os.path.isdir(source_path) else TYPE_FILE)

            # if score is not above zero, we skip this file
            if score <= 0:
                self.message('Skipping file with score {}: {}'.format(score, source_path), msglevel, level+1)
                continue

            ### if we get here, we need to copy the file ###

            # if a directory, recurse to it
            if os.path.isdir(source_path):
                self.message('Creating directory with score {}: {}'.format(score, source_path), msglevel, level+1)
                # create it in the destination and recurse
                if not os.path.exists(dest_path):
                    os.mkdir(dest_path)
                elif not os.path.isdir(dest_path):  # could be a file or link
                    os.unlink(dest_path)
                    os.mkdir(dest_path)
                self.copy_dir(source_path, dest_path, level+1)

            # if a regular Javscript file, run through the static file processors (scripts group)
            elif ext == '.js' and not self.options.get('no_minify') and jsmin:
                self.message('Including and minifying file with score {}: {}'.format(score, source_path), msglevel, level+1)
                with open(source_path, encoding=encoding) as fin:
                    with open(dest_path, 'w', encoding=encoding) as fout:
                        minified = minify(fin.read(), jsmin)
                        fout.write(minified)


            # if a CSS file, run through the static file processors (styles group)
            elif ext == '.css' and not self.options.get('no_minify') and cssmin:
                self.message('Including and minifying file with score {}: {}'.format(score, source_path), msglevel, level+1)
                with open(source_path, encoding=encoding) as fin:
                    with open(dest_path, 'w', encoding=encoding) as fout:
                        minified = minify(fin.read(), cssmin)
                        fout.write(minified)

            # otherwise, just copy the file
            else:
                self.message('Including file with score {}: {}'.format(score, source_path), msglevel, level+1)
                shutil.copy2(source_path, dest_path)



##############################################
###   Utility functions

def minify(text, minifier):
    '''Minifies the source text (if needed)'''
    # there really isn't a good way to know if a file is already minified.
    # our heuristic is if source is more than 50 bytes greater of dest OR
    # if a hard return is found in the first 50 chars, we assume it is not minified.
    minified = minifier(text)
    if  abs(len(text) - len(minified)) > 50 or '\n' in text[:50]:
        return minified
    return text
